print("import: 'mrjob'")
import mrjob

print("import: 'mrjob.examples'")
import mrjob.examples

print("import: 'mrjob.fs'")
import mrjob.fs

print("import: 'mrjob.logs'")
import mrjob.logs

print("import: 'mrjob.tools'")
import mrjob.tools

print("import: 'mrjob.tools.emr'")
import mrjob.tools.emr

